
"""Bohra 
.. moduleauthor:: Kristy Horan <kristyhoran15@gmail.com>
"""
__author__ = "Kristy Horan"
__license__ = "GPL"
__version__ = "1.0.1"
__maintainer__ = "Kristy Horan"
__email__ = "kristyhoran15@gmail.com"
__status__ = "Development"


import pathlib
import os, getpass, shutil, re, psutil
import pandas
import jinja2
import sh
import logging
import filecmp
import datetime
import numpy
import itertools
import subprocess
from Bio import SeqIO, Phylo
from packaging import version
from bohra.write_snakemake import MakeWorkflow
from bohra.write_report import Report


class RunSnpDetection(object):
    '''
    A class for Bohra pipeline object
    '''
    
    def __init__(self, args):
        # get date and time
        self.now = datetime.datetime.today().strftime("%d_%m_%y_%H")
        self.day = datetime.datetime.today().strftime("%d_%m_%y")
        # get the working directory
        self.workdir = pathlib.Path(args.workdir)
        # path to pipeline resources
        self.resources = pathlib.Path(args.resources)
        # path to reference and mask
        self.ref = pathlib.Path(args.reference)
        # 
        # (args.mask)
        if args.mask:
            rerun_core = self.check_mask(args.mask)
        else:
            self.mask = ''
        # path to input file
        if args.input_file == '':
            self.log_messages('warning', 'Input file can not be empty, please set -i path_to_input to try again')
            raise SystemExit()
        else:
            self.input_file = pathlib.Path(args.input_file)
        # # path to a source file for tracking of reference, jobid and mask
        # self.source_log_path = pathlib.Path(self.workdir, 'source.log')
        # job id
        self.job_id = self._name_exists(args.job_id)
        # other variables
        # min aln 
        self.minaln = args.minaln
        # set threads
        
        # user
        self.user = getpass.getuser()
        # gubbins TODO add back in later!!
        # if not args.gubbins:
        #     self.gubbins = numpy.nan
        # else:
        #     self.gubbins = args.gubbins
        
        self.gubbins = numpy.nan

        if isinstance(args.prefillpath, str):
            self.prefillpath = args.prefillpath
        elif args.mdu:
            self.prefillpath = '/home/seq/MDU/QC'
        else:
            self.prefillpath = ''
        self.force = args.force
        self.dryrun = args.dryrun
        self.pipeline = args.pipeline
        self.cpus = args.cpus
        
        self.assembler = args.assembler
        self.snippy_version = ''
        self.assembler_dict = {'shovill': 'shovill', 'skesa':'skesa','spades':'spades.py'}
        self.set_snakemake_jobs()

    def set_snakemake_jobs(self):
        '''
        set the number of jobs to run in parallel based on the number of cpus from args
        '''
        if int(self.cpus) < int(psutil.cpu_count()):
            self.jobs =  self.cpus
        else:
            self.jobs = 1

    def log_messages(self, type, message):
        '''
        Will log messages to the screen and also add them to job.log
        input:
            :type: type of message info or warning
            :message: message to log

        '''

        if type == 'warning':
            logging.warning(message)
            print(f"WARNING: {message}")
        if type == 'info':
            logging.info(message)
            # print(f"{message}")
    
    def force_overwrite(self):
        '''
        will force pipeline to run in an existing folder - removes isolate and source logs
        '''
        isolatelog = self.workdir / f"isolates.log"
        sourcelog = self.workdir / f"source.log"
        # joblog = self.workdir / f"job.log"
        if isolatelog.exists():
            isolatelog.unlink()
        if sourcelog.exists():
            sourcelog.unlink()
        # if joblog.exists():
        #     joblog.unlink()

        return False

    def check_setup_files(self):
        '''
        check that the working directory, resources directory and the input file exist
        '''
        self.path_exists(self.workdir, v = False) 
        self.path_exists(self.resources, v = False)
        self.path_exists(self.input_file)
    
    def check_snippy(self):
        '''
        check for snippy
        '''
        version_pat = re.compile(r'\bv?(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<release>[0-9]+)(?:\.(?P<build>[0-9]+))?\b')
        try:
            snippy = subprocess.run(['snippy', '--version'], stderr=subprocess.PIPE)
            snippy = snippy.stderr.decode().strip()
            self.snippy_version = version_pat.search(snippy)
            self.log_messages('info', f"Snippy {snippy} found. Good job!")
            
            return(version_pat.search(snippy))
        except FileNotFoundError:
            self.log_messages('warning', f"snippy is not installed.")
            raise SystemExit
    
    def check_snippycore(self):
        '''
        check for snippy-core
        '''
        self.check_installation('snippy-core')

    def check_snpdists(self):
        '''
        check for snp-dists
        '''
        self.check_installation('snp-dists')



    def check_iqtree(self):
        '''
        check iqtree
        '''
        self.check_installation('iqtree')

    def check_installation(self,software):
        '''
        Check that software is installed
        input:
            :software: the name of the software - must be command line name 
        '''

        if shutil.which(software):
            self.log_messages('info', f"{software} is installed")
        else:
            self.log_messages('warning', f"{software} is not installed, please check dependencies and try again.")
            raise SystemExit


    def check_assembler(self):
        '''
        check version of assembler
        '''
        ret = 0
        
        self.check_installation(self.assembler_dict[self.assembler])


    def check_assemble_accesories(self):
        '''
        check the assembly accessories mlst, kraken, abricate and prokka
        '''
        accessories = ['mlst', 'kraken2', 'abricate', 'prokka']
        
        for a in accessories:
            self.check_installation(a)
    
    def check_roary(self):
        '''
        check roary is installed
        '''

        self.check_installation('roary')


    def check_deps(self):
        '''
        check dependencies Snippy, snippy-core, snp-dists, iqtree
        '''
        # TODO check all software tools used and is there a way to check database last update??
        # TODO check assemblers
        self.log_messages('info', f"Checking software dependencies")
        if self.pipeline != 'a':
            self.check_snippycore()
            self.check_snpdists()
            self.check_iqtree()
            return(self.check_snippy())
        if self.pipeline != 's':
            self.check_assembler()
            self.check_assemble_accesories()
        if self.pipeline == 'all':
            self.check_roary()
        

    # def check_validation(self, validation_type):
    #     '''
    #     Check if this is a validation run 
    #     '''
    #         if validation_type == 'both':
    #             validate = ['snps', 'clusters']
    #         else:
    #             if validation_type in ['snps', 'clusters']:
    #                 validate = validation_type.split(',')
    #             else:
    #                 self.log_messages('warning', f"{validation_type} is not a valid input. Correct options can be taken from {' '.join(['snps', 'clusters', 'both'])}. Please try again")
    #                 raise SystemExit

    #         return(validate)
    
    

    def run_checks(self):
        '''
        Run checks prior to start - checking all software is installed, if this is a rerun and the input files
        '''
        self.check_setup_files()
        self.check_rerun()
        self.check_deps()
        # check reference
        if self.pipeline != 'a':
            if self.ref == '':
                self.log_messages('warning', f"You are trying call SNPs, a reference file is required. Please try again using '-r path to reference'")
                raise SystemExit
            else:
                self.ref = self.link_file(self.ref)
        
            

    def set_source_log(self):
        '''
        set the reference, mask and id for tracking and potential.
        
            
        '''   
        new_df = pandas.DataFrame({'JobID':self.job_id, 'Reference':f"{self.ref}",'Mask':f"{self.mask}", 
                                    'MinAln':self.minaln, 'Pipeline': self.pipeline, 'CPUS': self.cpus, 'Assembler':self.assembler,
                                    'Gubbins': self.gubbins, 'Date':self.day, 'User':self.user, 'snippy_version':self.snippy_version, 'input_file':f"{self.input_file}",'prefillpath': self.prefillpath}, 
                                    index=[0], )

        source_path = self.workdir / 'source.log'
        if source_path.exists():
            source_df = pandas.read_csv(source_path, '\t')
            source_df = source_df.append(new_df)
        else:
            source_df = new_df
        
        source_df.to_csv(source_path , index=False, sep = '\t')

    def check_rerun(self):
        '''
        CHeck if the job is a rerun of an existing job, if so print message informing user and exit

        '''

        source_path = self.workdir / 'source.log'
        # if the path is a string convert to Path
        if isinstance(source_path, str):
            source_path = pathlib.Path(source_path)
        if source_path.exists():
            self.log_messages('warning',f"This may be a re-run of an existing job. Please try again using rerun instead of detect OR use -f to force an overwrite of the existing job.")
            self.log_messages('warning',f"Exiting....")
            self.log_messages('info',f"{60 * '='}")
            raise SystemExit()
        else:
            return False


    def path_exists(self,path, v = True):
        '''
        ensure files are present, if so continues if not quits with FileNotFoundError
        input:
            :path: patht to files for pipeline
            :v: if v == True print message, else just check
        output:
            returns True (or fails with FileNotFoundError)
        '''
        
        if not path.exists():
            self.log_messages('warning', f"The {path.name} does not exist.")
            raise FileNotFoundError(f"{path.name}")
        else:
            if v == True:
                self.log_messages('info', f"Found {path.name}.")

            return True

    
        


    def _name_exists(self, name):
        '''
        check if the name is an empty string JOB id can not be empty
       
        '''
        
        if isinstance(name, str):
            if len(name) == 0:
                self.log_messages('warning', 'Job id ca not be empty, please set -j job_id to try again')
                raise SystemExit()
            else:
                return name
        else:
            self.log_messages('warning', 'Job id ca not be empty, please set -j job_id to try again')
            raise SystemExit()

    def link_reads(self, read_source, isolate_id, r_pair):
        '''
        check if read source exists if so check if target exists - if not create isolate dir and link. If already exists report that a dupilcation may have occured in input and procedd

        '''
        # check that job directory exists
        J = pathlib.Path(self.workdir, self.job_id)
        if not J.exists():
            J.mkdir()
        # check that READS exists
        R = J / 'READS'
        if not R.exists():
            R.mkdir()
        
        # check that source is path
        if isinstance(read_source, str):
            read_source = pathlib.Path(read_source)
        # check that source exists
        if read_source.exists():
            I = R / f"{isolate_id}" # the directory where reads will be stored for the isolate
            if not I.exists():
                I.mkdir()
            read_target = I / f"{r_pair}"
            if not read_target.exists():
                read_target.symlink_to(read_source)

    def unzip_files(self,path, unzipped):
        '''
        if a zipped reference is provided try to unzip and then return the unzipped pathname. If unable to unzip then supply message and exit
        input:
            :path: pathname  of file to unzip string
            :unzipped: unzipped path
        '''
        
        cmd = f"gzip -d -c {path} > {p.stem}"
        try:
            subprocess.run(cmd, shell = True)
            return pathlib.Path(p.stem)
        except:
            self.log_messages('warning', f"{path} can not be unzipped. This may be due to file permissions, please provide path to either an uncompressed reference or a file you have permissions to.")
            raise SystemExit            

        


    def link_file(self, path):
        '''
        check if file exists and copy to workingdir
        input:
            :path: path to file
        if path does not exist then return false (calling function will request path). 
        if it does exist, then create symlink to the working dir 
        output:
            returns path.name (str)   
        '''
        found = False
        while not found: 
            
            if path.exists():
                target = self.workdir / path.name
                # use rename to copy reference to working directory
                # if the reference is not already in the working directory symlink it to working dir
                if not target.exists():
                    logging.info(f"Linking {path.name} to {self.workdir.name}")
                    target.symlink_to(path)
                    if path.suffix == '.gz':
                        path = self.unzip_files(path.name, path.name.strip('.gz'))
                    # TODO add in unzip option unzip 
                
                found = True
            else:
                self.log_messages('warning', f"Path to {path} does not exist. Please provide a valid path to a file and try again")
                raise SystemExit
                # path = pathlib.Path(path)
        
        return  path.name

    def check_mask(self, mask, original_mask = False):
        '''
        input:
            :mask: path to mask file (str)
            :original_mask: name of orignal mask to use in rerun
        output:
            :mask: path to mask  file in workingdir (str) and a boolean True == rerun snippy-core, tree and distance, False == no need to rerun snippy-core and tree
        '''
        
        # if there is a file path added the generate a symlink
        if len(mask) > 0:
                m = pathlib.Path(mask)
                if f"{m.name}" == original_mask:
                    self.mask = original_mask
                    return original_mask
                else:
                    m = self.link_file(m)
                    self.mask = m
                    return m
        elif len(mask) == 0 and original_mask:
            self.mask = original_mask
            return original_mask
        else:
            return ''
    
    def min_four_samples(self, tab):
        '''
        Ensure that there are a minimum of four samples
        returns True if four or more samples
        '''
        return tab.shape[0] < 4

    def three_cols(self, tab):
        '''
        Ensure that there are 3 columns, isolate, R1 and R2
        returns True if 3 columns False otherwise
        
        '''
        if tab.shape[1] == 3:
            return True
        else:
            return False

    def all_data_filled(self, tab):
        '''
        Ensure that all fields contain data - no NA's
        returns True if there are no nan, False otherwise
        '''
        return tab.isnull().sum().sum() == 0
    

    def check_input_structure(self, tab):
        '''
        check that the structure of the input file is correct, 3 columns, with a minimum of 4 isolates
        :input
            :tab: dataframe of the input file
        :output
            True if make it to the end without exiting with a TypeWarning the user of the reason file was rejected
        '''
        # if the structure of the file is incorrect tell user and kill process
        # not the right information (not three columns)
        
        if not self.three_cols(tab):
            self.log_messages('warning',f"{self.input_file} does not appear to be in the correct configuration")
            raise TypeError(f"{self.input_file} has incorrect number of columns")
        # if there are not enough isolates (>4)
        
        if self.min_four_samples(tab):
            self.log_messages('warning', f"{self.input_file} does not contain enough isolates. The minimum is 4.")
            raise TypeError(f"{self.input_file} has incorrect number of isolates")
        # if any na present indicates that not the full info has been provided
        if not self.all_data_filled(tab):
            self.log_messages('warning',f"{self.input_file} appears to be missing some inforamtion.")
            raise TypeError(f"{self.input_file} appears to be missing some inforamtion.")
        
        return True

    def check_reads_exists(self, tab):
        '''
        check that the correct paths have been given
        if reads not present path_exists will cause a FileNotFound error and warn user
        :input
            :tab: dataframe of the input file
        '''
        for i in tab.itertuples():
            
            if not '#' in i[1]:
                r1 = i[2]
                r2 = i[3]
                self.path_exists(pathlib.Path(r1), v = False)
                self.link_reads(pathlib.Path(r1), isolate_id=f"{i[1].strip()}", r_pair='R1.fq.gz')
                self.path_exists(pathlib.Path(r2), v = False)
                self.link_reads(pathlib.Path(r2), isolate_id=f"{i[1].strip()}", r_pair='R2.fq.gz')
        return True

    def set_isolate_log(self, tab, logfile, validation = False):
        '''
        add the isolates to a log file also adds in anoterh check that this is a rerun of an existing job
        input:
            :tab: dataframe of the isolates to add - same structure as original input file, but not needing to be > 4 isolate
            :logfile: path to logfile
            
        '''        
        self.check_input_structure(tab=tab)
        self.check_reads_exists(tab=tab)
        
        lf = pandas.DataFrame({'Isolate': [i for i in list(tab.iloc[ : , 0]) if '#' not in i ], 'Status': f"INCLUDED", 'Date': self.day})
        lf['Status'] = numpy.where(lf['Isolate'].str.contains('#'), f"REMOVED", lf['Status'])
        isolates = lf[lf['Status'].isin(['INCLUDED', 'ADDED'])]['Isolate']
        lf.to_csv(logfile, sep = '\t', index = False)
        return list(isolates)
        
    
    def set_workflow_input(self, validation = False):
        '''
        read the input file and check that it is the correct format
        input:
            
        output:
            a list of isolates that will be used in generation of job configfile.
        '''
        logfile = pathlib.Path(self.workdir, 'isolates.log') 
        # make df of input file
        

        # the path to the user provided input file check if it exists
        # read the file
        
        tab = pandas.read_csv(self.input_file, sep = None, engine = 'python', header = None)
        
        # self.log_messages('info', f"input file has been opened as a dataframe")
        isolates = self.set_isolate_log(tab = tab, logfile = logfile, validation = validation)
        
        # get a list of isolates
        return(list(set(isolates))) 
        
    def write_pipeline_job(self, maskstring,  script_path = f"{pathlib.Path(__file__).parent}", resource_path = f"{pathlib.Path(__file__).parent / 'templates'}"):
        '''
        write out the pipeline string for transfer to job specific pipeline
        '''
        wd = self.workdir / self.job_id
        
        c = MakeWorkflow()
        config_params = c.write_config_params()
        # config_params = config_params.write_config_params()

        p = MakeWorkflow()
        self.log_messages('message', f"writing the pipeline for {self.job_id}")
        
        # pipeline always has seqdata in 
        pipelinelist = [p.write_all(self.pipeline), p.write_seqdata(prefillpath = self.prefillpath),p.write_combine_seqdata()]

        # for just snps contains snippy, snippy-core, dists and tree
        snps_pipeline = [p.write_snippy(), p.write_qc_snippy_initial(), p.write_snippy_core(mask = maskstring), p.write_snp_dists(), p.write_tree(script_path=script_path, alntype='core')]
        # for just assemblies
        assembly_pipeline = [p.write_assemblies(prefillpath = self.prefillpath), p.write_resistome(), p.write_mlst(),p.write_kraken(prefillpath = self.prefillpath), p.write_combine(), p.write_assembly_stats(), p.write_prokka(), p.write_gff_summary(), p.write_combine_kraken()]
        # for roary
        roary_pipeline = [p.write_roary(), p.write_pan_graph()]

        # pipeline can always ends with report
        report_pipeline = [p.write_report_collation(pipeline = self.pipeline), p.write_html(pipeline = self.pipeline,workdir = self.workdir, resources = resource_path, job_id = self.job_id, script_path = script_path, assembler = self.assembler_dict[self.assembler])]
        
        if self.pipeline == 's':
            pipelinelist.extend(snps_pipeline)
        elif self.pipeline == 'a':
            pipelinelist.extend(assembly_pipeline)
        elif self.pipeline == 'sa':
            pipelinelist.extend(snps_pipeline)
            pipelinelist.extend(assembly_pipeline)
        elif self.pipeline == 'all':
            pipelinelist.extend(snps_pipeline)
            pipelinelist.extend(assembly_pipeline)
            pipelinelist.extend(roary_pipeline)
        pipelinelist.extend(report_pipeline)


        return(wd, config_params, '\n'.join(pipelinelist))


    def setup_workflow(self, isolates, config_name = 'config.yaml', snake_name = 'Snakefile'):
        '''
        generate job specific snakefile and config.yaml
        input:
            :isolates: a list of isolates that need to be included
        '''

        self.log_messages('info',f"Setting up {self.job_id} specific workflow")
        

        if self.gubbins == True:
            gubbins_string = f"""
        'gubbins.aln', 'gubbins.treefile'
            """
        else:
            gubbins_string = " "
        # make a masking string

        if self.mask != '':
            maskstring = f"--mask {self.workdir / self.mask}"
        else:
            maskstring = ''
                
        # read the config file which is written with jinja2 placeholders (like django template language)
        config_template = jinja2.Template(pathlib.Path(self.resources, 'config_snippy.yaml').read_text())
        config = self.workdir / f"{self.job_id}"/ f"{config_name}"
        
        config.write_text(config_template.render(reference = f"{pathlib.Path(self.workdir, self.ref)}", cpus = self.cpus, name = self.job_id,  minperc = self.minaln,now = self.now, maskstring = maskstring, day = self.day, isolates = ' '.join(isolates)))
        
        self.log_messages('info',f"Config file successfully created")

        snk_template = jinja2.Template(pathlib.Path(self.resources, 'Snakefile_base').read_text())
        snk = self.workdir / snake_name

        # write out custom parts of pipeline
        wd, config_params, pipelinestring = self.write_pipeline_job(maskstring = maskstring)
        snk.write_text(snk_template.render(configfile = f"configfile: '{config_name}'",config_params = config_params, pipeline = pipelinestring, workdir=f"workdir: '{wd}'", gubbins_input = gubbins_string)) 
        
        self.log_messages('info',f"Snakefile successfully created")

 
    def run_workflow(self,snake_name = 'Snakefile'):
        '''
        run snp_detection
        set the current directory to working dir for correct running of pipeline
        if the pipeline wroks, return True else False
        '''
        if self.force:
            force = f"-F"
        else:
            force = f""
        os.chdir(self.workdir)
        if self.dryrun:
            cmd = f"snakemake -np -s {snake_name} 2>&1 | tee -a job.log"
        else:
            cmd = f"snakemake -s {snake_name} -j {self.jobs} {force} 2>&1 | tee -a job.log"
            # cmd = f"snakemake -s {snake_name} --cores {self.cpus} {force} "
        wkf = subprocess.run(cmd, shell = True)
        if wkf.returncode == 0:
            return True
        else:
            return False



    def run_pipeline(self):
        '''
        run pipeline, if workflow runs to completion print out a thank you message.
        '''
        # if -f true force a restart
        if self.force:
            self.force_overwrite()
        # check the pipeline setup 
        self.run_checks()
        # update source data in source.log
        self.set_source_log()
        # open the input file and check it is in the minimal correct format 
        isolates = self.set_workflow_input()
        # setup the workflow files Snakefile and config file
        self.setup_workflow(isolates = isolates)
        # run the workflow
        if self.run_workflow():
            # TODO add in cleanup function to remove snakemkae fluff 
            if not self.dryrun:
                self.log_messages('info', f"Report can be found in {self.job_id}")
                self.log_messages('info', f"Process specific log files can be found in process directories. Job settings can be found in source.log") 
            else:
                if self.force:
                    force = f"-F"
                self.log_messages('info', f"snakemake -j {self.jobs} {force} 2>&1 | tee -a job.log")
            self.log_messages('info', f"Have a nice day. Come back soon.") 
            self.log_messages('info',f"{60 * '='}")




import sys, pathlib, pandas, pytest, numpy

from unittest.mock import patch

from bohra.SnpDetection import RunSnpDetection
from bohra.ReRunSnpDetection import ReRunSnpDetection




def test_name_string():
        '''
        assert true when the input is a string > len 0
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                assert detect_obj._name_exists('somestring')

def test_name_non_string():
        '''
        if a non string input is used return false
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                with pytest.raises(SystemExit):
                        detect_obj._name_exists(9)

def test_name_empty_string():
        '''
        confirm that False is returned if a 0 length strin is input
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                with pytest.raises(SystemExit):
                        detect_obj._name_exists('')
              

def test_3col_dimensions():
        '''
        return True when correct number of columns
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1], 'B':[2], 'C':[3]})
                assert detect_obj.three_cols(tab)

def test_2col_dimensions():
        '''
        return False when wrong number of columns present
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1], 'B':[2]})
                assert detect_obj.three_cols(tab) == False


def test_four_isolates():
        '''
        confirm that min of 4 rows are present
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1,2,3,4]})
                assert detect_obj.min_four_samples(tab) == False

def test_notfour_isolates():
        '''
        return false when less than 4 rows are present
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1,2,3]})
                assert detect_obj.min_four_samples(tab)

def test_missing():
        '''
        a full dataframe returns true
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1,2,3,4], 'B':[5,6,7,8], 'C':[9,10,11,12]})
                assert detect_obj.all_data_filled(tab)

def test_with_missing():
        '''
        return False if missing data present
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1,2,4, numpy.nan], 'B':[5,6,7,8], 'C':[9,10,11,12]})
                assert detect_obj.all_data_filled(tab) == False


def test_structure():
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                detect_obj = RunSnpDetection()
                tab = pandas.DataFrame({'A':[1,2,3,4], 'B':[5,6,7,8], 'C':[9,10,11,12]})
                assert detect_obj.check_input_structure(tab) == True


def test_path_exists():
        '''
        test that path_exists returns True
        '''
        with patch.object(RunSnpDetection, "__init__", lambda x: None):
                p = pathlib.Path('test', 'test_file.txt')
                detect_obj = RunSnpDetection()
                assert detect_obj.path_exists(p)




import pathlib
import os,getpass
import pandas
import jinja2
import sh
import logging
import filecmp
import datetime
import numpy
import itertools
import subprocess
import re
from packaging import version
from Bio import SeqIO, Phylo
from bohra.SnpDetection import RunSnpDetection

class ReRunSnpDetection(RunSnpDetection):
    '''
    A class for a Bohra reurn objects - inherits RunSnpDetection
    '''

    def __init__(self, args):
        # get date and time
        self.now = datetime.datetime.today().strftime("%d_%m_%y_%H")
        self.day = datetime.datetime.today().strftime("%d_%m_%y")
        # get the working directory
        self.workdir = pathlib.Path(args.workdir)
        # path to pipeline resources
        self.resources = pathlib.Path(args.resources)
        # path to reference => if args.reference is a string, check that it matches existing
        # set force based on args.. this will be set to true if ref is different and/or snippy version 
        self.force = False
        
        # get original data 
        self.get_source()
        # Reference mask and snippy
        
        if self.pipeline != 'a':
            self.check_reference(new = args.reference)
            # check dependencies
            self.check_for_snippy()
            self.mask = self.check_mask(args.mask, original_mask = self.original_mask)
        # user
        self.user = getpass.getuser()
        # gubbins TODO add back in later!!
        if not args.gubbins:
            self.gubbins = numpy.nan
        else:
            self.gubbins = args.gubbins
        
        self.dryrun = args.dryrun
        self.keep = args.keep
        
        self.assembler_dict = {'shovill': 'shovill', 'skesa':'skesa','spades':'spades.py'}

        self.cpus = args.cpus
        self.set_snakemake_jobs()

    def get_source(self):

        '''
        open the source.log file and extract reference, mask, date, snippy_version
        '''

        version_pat = re.compile(r'\bv?(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<release>[0-9]+)(?:\.(?P<build>[0-9]+))?\b')
        df = pandas.read_csv('source.log', sep = None, engine = 'python')
        self.pipeline = df.loc[df.index[-1], 'Pipeline']
        if self.pipeline != 'a':
            self.original_reference = df.loc[df.index[-1], 'Reference']
            self.original_mask = df.loc[df.index[-1],'Mask'] if isinstance(df.loc[df.index[-1],'Mask'], str) else ''
            self.original_snippy_version = version_pat.search(df.loc[df.index[-1], 'snippy_version'])
        if self.pipeline != 's':
            self.assembler = df.loc[df.index[-1], 'Assembler']
        self.orignal_date = df.loc[df.index[-1], 'Date']
        self.input_file = pathlib.Path(f"{df.loc[df.index[-1], 'input_file']}")
        # print(self.input_file)
        self.job_id = df.loc[df.index[-1], 'JobID']
        self.cpus = df.loc[df.index[-1], 'CPUS']
        self.prefillpath = df.loc[df.index[-1], 'prefillpath']
        self.minaln = df.loc[df.index[-1], 'MinAln']
        
        # return reference, mask, snippy_version, date, input_file, pipeline
        

    def check_reference(self, new):
        '''
        Check that reference used in rerun is the same as the reference used in previous run. If not will need to force new SNP detection
        '''
        # check if refs are the same if not set self.ref to new and change to force, else set ref to original
        if isinstance(new, str) and len(new) > 0 and len(self.original_reference) > 0:
            new_reference = pathlib.Path(f"{new}")
            if f"{new_reference.name}" == self.original_reference:
                self.ref = pathlib.Path(self.original_reference)
            
            else:
                self.ref = self.link_file(path = new_reference)
                self.force = True
                self.log_messages('message', f"You have chosen a different reference from the previous run. Snippy will be forced to run again from the beginning.")
        elif isinstance(new, str) and len(new) == 0 and len(self.original_reference) > 0:
            self.ref = pathlib.Path(self.original_reference)
        else:
            self.log_messages('message', 'There appears to be something wrong with the reference. You will need to run Bohra using the run command.')


    def check_for_snippy(self):
        '''
        Check the version of Snippy, is different will need to force new SNP detection
        '''
        self.check_setup_files()
        self.current_snippy_version = self.check_deps()
        if self.current_snippy_version.group("major", "minor") != self.original_snippy_version.group("major", "minor"):
            self.force = True
            self.log_messages('info', f"You are using a different version of Snippy for this re-run, SNP calling will be repeated.")


    def update_source_log(self):
        '''
        update source.log if user changes parameters
        '''
        df = pandas.read_csv('source.log', sep = None, engine = 'python')
        data =pandas.DataFrame({'JobID':self.job_id, 'Reference':self.ref,'Mask':self.mask, 'Pipeline': self.pipeline, 'CPUS': self.cpus,'MinAln':self.minaln,'Gubbins': self.gubbins, 'Date':self.day, 'User':self.user,'snippy_version':self.current_snippy_version ,'input_file':f"{self.input_file}",'prefillpath': self.prefillpath,'Assembler':self.assembler},index=[0])
        df = df.append(data)
        df.to_csv('source.log', index=False, sep = '\t')
    
        
    # def run_with_gubbins(self):
    #     '''
    #     Check if user wants to use Gubbins - return Y or N
    #     '''
    #     if self.confirm(f"Would you like to use gubbins to remove recombination?", True):
    #         self.log_messages('info', f"Gubbins will be used to remove recombination today.")
    #         return True
    #     else:
    #         self.log_messages('info', f"No recombination filtering will be used.")
    #         return False
    # TODO change this delete to unlink from pathlib
    
    def rerun_report(self):
        '''
        Remove report directory from previous run 
        '''
        # os.chdir(self.workdir)
        p1 = pathlib.Path(self.workdir, self.job_id, 'report')
        p2 = pathlib.Path(self.workdir, self.job_id, f"report_{self.orignal_date}")
        if self.keep:
            cmd = f"mv {p1} {p2}"
        else:
            cmd = f"if [ -d {p1} ];then rm -r {p1}; fi"
        subprocess.run(cmd, shell = True)
    
    def remove_core(self):
        '''
        Need to remove core_isolates.txt to get snakemake to redo snippy core step
        '''
        corefiles = sorted(pathlib.Path(self.workdir, self.job_id).glob('core*'))
        if corefiles:
            for core in corefiles:
                core.unlink()
        
        

    def run_pipeline(self):
        '''
        Rerun the pipeline
        '''
        # update source
        self.update_source_log()
        self.rerun_report()
        self.remove_core()

        isolates = self.set_workflow_input()
        # setup the workflow files Snakefile and config file
        self.setup_workflow(isolates = isolates)
        # run the workflow
        if self.run_workflow(): 
            if not self.dryrun:
                self.log_messages('info', f"Report can be found in {self.job_id}")
                self.log_messages('info', f"Process specific log files can be found in process directories. Job settings can be found in source.log") 
    
            self.log_messages('info', f"Have a nice day. Come back soon.") 
            self.log_messages('info',f"{60 * '='}")



class MakeWorkflow():
	'''
	A Class to write out the snakemkae file for each specific job

	'''

	# TODO add in logic for species and exp for validation and prefill path
	def write_config_params(self, rerun = False):
		
		return(f"""
job = config['name']
min_aln = int(config['min_perc'])
now = config['now']
REFERENCE = config['reference']


""")

	
	def write_all(self, pipeline = 'sa'):
		base = f"""	expand(\"{{sample}}/seqdata.tab\", sample = SAMPLE),
		\"report/seqdata.tab\""""

		sstring = f"""
		expand(\"{{sample}}/snps.vcf\", sample = SAMPLE),
 		expand(\"{{sample}}/snps.aligned.fa\", sample = SAMPLE),
		\"core.vcf\", 
		\"distances.tab\",
		\"core.treefile\", 
		\"report/core_genome.tab\", 
		\"report/core.treefile\", 
		\"report/distances.tab\",
		\"report/core.tab\""""

		astring = f"""
		expand(\"{{sample}}/{{sample}}.fa\", sample = SAMPLE),
		expand(\"{{sample}}/kraken.tab\", sample = SAMPLE),
		expand(\"{{sample}}/resistome.tab\", sample = SAMPLE),
		expand(\"prokka/{{sample}}/{{sample}}.gff\", sample = SAMPLE),
		expand(\"prokka/{{sample}}/{{sample}}.txt\", sample = SAMPLE),
		\"mlst.tab\", 
		\"species_identification.tab\",
		\"denovo.tab\", 
		\"assembly.tab\", 
		\"resistome.tab\",
		\"report/assembly.tab\",
		\"report/mlst.tab\", 
		\"report/species_identification.tab\", 
		\"report/resistome.tab\""""

		rstring= f"""
		\"roary/gene_presence_absence.csv\", 
		\"pan_genome.svg\", 
		\"report/pan_genome.svg\",
		\"report/summary_statistics.txt\""""

		report_string = f"""
		\"report/report.html\""""
		i_dict = {
			's': f"""{base},{sstring},{report_string}
""",
			'a': f"""{base},{astring},{report_string}
""",
			'sa': f"""{base},{astring},{sstring},{report_string}
""",
			'all': f"""
{base},{astring},{sstring},{rstring},{report_string}
"""
		}

		return(f"""
rule all:
	input:
	{i_dict[pipeline]}
""")


	def write_seqdata(self, prefillpath = False):
		return(f"""
rule seqdata:
	input:
		'READS/{{sample}}/R1.fq.gz',
		'READS/{{sample}}/R2.fq.gz'
	output:
		"{{sample}}/seqdata.tab"
	shell:
		\"""
		YIELDPATH={prefillpath}/{{wildcards.sample}}/yield.tab
		if [ -f $YIELDPATH ]; then
			cp $YIELDPATH {{output}}
		else
			fq --quiet --ref {{REFERENCE}} {{input[0]}} {{input[1]}} > {{output}}
		fi
		\"""
""")

	def write_kraken(self, prefillpath = ''):
		return(f"""
rule kraken:
	input:
		'READS/{{sample}}/R1.fq.gz',
		'READS/{{sample}}/R2.fq.gz'
	output:
		"{{sample}}/kraken.tab"

	shell:
		\"""
		KRAKENPATH={prefillpath}/{{wildcards.sample}}/kraken2.tab
		if [ -f $KRAKENPATH ]; then
			cp $KRAKENPATH {{output}}
		else
			kraken2 --paired {{input[0]}} {{input[1]}} --memory-mapping --minimum-base-quality 13 --report {{output}} 
		fi
		\"""
		
""")

	def write_combine_kraken(self):
		return(f"""
rule combine_kraken:
	input: 
		expand("{{sample}}/kraken.tab", sample = SAMPLE)
	output:
		"species_identification.tab"
	run:
		import pandas, pathlib, subprocess
		kfiles = f"{{input}}".split()
		id_table = pandas.DataFrame()
		for k in kfiles:
			kraken = pathlib.Path(k)
			
			df = pandas.read_csv(kraken, sep = "\t", header =None, names =  ['percentage', 'frag1', 'frag2','code','taxon','name'])
			df['percentage'] = df['percentage'].apply(lambda x:float(x.strip('%')) if isinstance(x, str) == True else float(x)) #remove % from columns
			df = df.sort_values(by = ['percentage'], ascending = False)
			df = df[df['code'].isin(['U','S'])]     
			df = df.reset_index(drop = True) 
			tempdf = pandas.DataFrame()
			d = {{'Isolate': f"{{kraken.parts[0]}}",    
					'#1 Match': df.ix[0,'name'].strip(), '%1': df.ix[0,'percentage'],
					'#2 Match': df.ix[1,'name'].strip(), '%2': df.ix[1,'percentage'],       
					'#3 Match': df.ix[2,'name'].strip(), '%3': df.ix[2,'percentage'] ,
					'#4 Match': df.ix[3,'name'].strip(), '%4': df.ix[3,'percentage']
					}}
			tempdf = pandas.DataFrame(data = d, index= [0])
			if id_table.empty:
					id_table = tempdf
			else:
					id_table = id_table.append(tempdf)
		id_table.to_csv(f"{{output}}", sep = "\t", index = False)
		subprocess.run(f"sed -i 's/%[0-9]/%/g' {{output}}", shell=True)
""")

	def write_combine_seqdata(self):
		return(f"""
rule combine_seqdata:
	input:
		expand("{{sample}}/seqdata.tab", sample = SAMPLE)
	output:
		"seqdata.tab"
	run:
		import pathlib, pandas, numpy
		sdfiles = f"{{input}}".split()
		seqdata = pandas.DataFrame()
		for sd in sdfiles:

			p = pathlib.Path(sd)
			df = pandas.read_csv(sd, sep = "\t")
			df.columns = ['Isolate', f"{{p.parts[0]}}"]
			#print(df)
			dft = df.transpose()
			dft.columns = dft.iloc[0]
			dft = dft.drop(['Isolate'])
			#print(dft)
			#print(dft.index)
			depth = dft['Depth'].str.strip('x')
			print(depth)
			dft['Depth'] = int(depth)
			if seqdata.empty:
				seqdata = dft
			else:
				seqdata = seqdata.append(dft)
		seqdata['Quality'] = numpy.where(seqdata['Depth'] >= 40, 'PASS','FAIL')
		seqdata = seqdata.reset_index()
		seqdata = seqdata.rename(columns={{seqdata.columns[0]: 'Isolate'}})
		seqdata.to_csv(f"{{output}}", sep = '\t', index = False)
		

""")

	def write_snippy(self):
		
		return(f"""
rule snippy:
	input:
		'READS/{{sample}}/R1.fq.gz',
		'READS/{{sample}}/R2.fq.gz'
	output:
		'{{sample}}/snps.vcf',
		'{{sample}}/snps.aligned.fa'
	threads:
		8
	shell:
		\"""
		snippy --outdir {{wildcards.sample}} --ref {{REFERENCE}} --R1 {{input[0]}} --R2 {{input[1]}} --force --cpus {{threads}}
		\"""
	""")


	def write_qc_snippy_initial(self):
		return(f"""
rule qc_snippy: 
	input:
		expand('{{sample}}/snps.aligned.fa', sample = SAMPLE)
		
	output:
		'core_isolates.txt'
		
	run:
		from Bio import SeqIO
		import pathlib
		import pandas
		# create an output
		isolate_list = []
		excluded_list = []
		outfile = pathlib.Path(f"{{output[0]}}")
		# get input file list
		input_list = f"{{input}}".split()
		# set the log path
		logpath = pathlib.Path('isolates.log')
		for i in input_list: # for each input file
			# get the isolate name
			p = pathlib.Path(f"{{i}}")
			isolate = p.parts[-2]
			if p.exists(): # if the file exists open it
				fasta = p.open()
				for i in SeqIO.parse(fasta,'fasta'): # use BioPython to determine percent alignment
					length = len(i.seq)
					nocov = i.seq.count('-')
					lowcov = i.seq.count('N')
					het = i.seq.count('n')
					unaln = nocov + lowcov + het
					perc_aln = 100*(length - unaln) / length
					# if the percent alignement is greater than the min alignment
					if perc_aln > min_aln:
						isolate_list.append(f"{{isolate}}")
					else:
						excluded_list.append(isolate)
						print(f"{{isolate}} has been excluded from the analysis due to poor alignement with reference")
						
		isolate_list = list(set(isolate_list))
		with open(outfile, 'w') as f:
			f.write('\\n'.join(isolate_list))
		# get log if the excluded list has any isolates in
		if excluded_list != []:
			if logpath.exists():
				lf = pandas.read_csv(logpath, sep = '\t', index_col = False)
				for e in excluded_list:
					lf.loc[lf['Isolate'] == e.strip('#'), 'Status'] = f"(FAILED ALIGNMENT (<{{min_aln}}% ALIGNMENT))"
					lf.loc[lf['Isolate'] == e.strip('#'), 'Date'] = f"{{config['day']}}"
					lf.to_csv(logpath, sep = '\t', index=False)

	""")


	def write_snippy_core(self, mask):
			return(f"""
rule run_snippy_core:
	input:
		'core_isolates.txt'
	output:
		'core.vcf',
		'core.txt',
		'core.aln', 
		'core.full.aln',
		'core.tab'
	
	shell:
		\"""
		snippy-core --ref {{REFERENCE}} {mask} $(cat core_isolates.txt)
		
		\"""
	""")

	def write_snp_dists(self):
		return(f"""
rule run_snpdists:
	input:
		'core.aln' 
	output:
		'distances.tab' 
	
	shell:
		\"""
		snp-dists {{input}} > {{output}}
		\"""
	""")


	def write_tree(self, script_path, alntype):
		return(f"""
rule calculate_iqtree_command_{alntype}:
	input:
		'{alntype}.aln',
		REFERENCE
	output:
		'run_iqtree_{alntype}.sh'
	run:
		from Bio import SeqIO
		import pathlib
		ref = pathlib.Path(REFERENCE)
		name = ref.stem
		if '.fa' not in REFERENCE:
			print(f"converting {{REFERENCE}}")
			SeqIO.convert(f"{{input[1]}}", 'genbank', f"{{name}}.fa", 'fasta')
			ref = f"{{name}}.fa"
		else:
			ref = f"{{ref}}"
		
		shell("bash {script_path}/iqtree_generator.sh {{ref}} {{input[0]}} {alntype} 20 > {{output}}")

rule run_iqtree_{alntype}:
	input:
		'run_iqtree_{alntype}.sh'
	
	output:
		'{alntype}.iqtree',
		'{alntype}.treefile',
		
	
	shell:
		\"""	
		bash run_iqtree_{alntype}.sh
		
		rm *.ckp.gz *.contree *.bionj
		\"""
		
	""")



	def write_assemblies(self, prefillpath = '', assembler = 'skesa'):
		
		if assembler == 'skesa':
			assemble = f"skesa --fastq {{input[0]}},{{input[1]}} --vector_percent 1 --use_paired_ends --cores {{threads}} > {{output}}"
		elif assembler == 'shovill':
			assemble = f""""
			shovill --outdir {{wildcards.sample}} --R1 {{input[0]}} --R2 {{input[1]}} --force --minlen 500 --cpus {{threads}}
			mv {{wildcards.sample}}/contigs.fa {{output}}
"""
		elif assembler == 'spades':
			assemble = f"""
			spades.py -o {{wildcards.sample}} -1 {{input[0]}} -2 {{input[1]}} --threads {{threads}}
			mv {{wildcards.sample}}/contigs.fasta {{output}}
"""
		
		return(f"""
rule assemble:
	input:
		'READS/{{sample}}/R1.fq.gz',
		'READS/{{sample}}/R2.fq.gz'
	output:
		'{{sample}}/{{sample}}.fa'
	threads:
		16
	shell:
		\"""
		ASSEMBLEPATH={prefillpath}/{{wildcards.sample}}
		if [ -f $ASSEMBLEPATH/contigs.fa ]; then
			cp $ASSEMBLEPATH/contigs.fa {{output}}

		else
			echo No assembly found. Assembling {{wildcards.sample}} with {assembler}
			{assemble}
		fi		
		\"""
	""") # TODO Will need to add in assembly options here

	def write_mlst(self):
		return(f"""
rule mlst:
	input:
		expand('{{sample}}/{{sample}}.fa', sample = SAMPLE)
	output:
		'mlst.tab'
	
	shell:
		\"""
		mlst --nopath {{input}} | sed 's/\.fa//g' | sed '1iIsolate\tScheme\tST\tAlleles' > {{output}}
		\"""

	""")

	def write_resistome(self):
		return(f"""
rule resistome:
	input:
		'{{sample}}/{{sample}}.fa'
	output:
		'{{sample}}/resistome.tab'
	
	shell:
		\"""
		abricate --nopath {{input}} > {{output}}
		\"""

	""")


	def write_combine(self):
		return(f"""
rule combine_results:
	input:
		expand('{{sample}}/resistome.tab', sample = SAMPLE)
		
	output:
		'resistome.tab'
		
	
	shell:
		\"""
		abricate --summary {{input}} | sed 's/\/resistome.tab//g' | sed 's/\#FILE/Isolate/g' > {{output[0]}}
		
		\"""
	""")

	def write_prokka(self):
		return(f"""
rule run_prokka:
	input:
		"{{sample}}/{{sample}}.fa"
	output:
		"prokka/{{sample}}/{{sample}}.gff","prokka/{{sample}}/{{sample}}.txt"
	shell:
		\"""
		prokka --outdir prokka/{{wildcards.sample}} --prefix {{wildcards.sample}} --mincontiglen 500 --notrna --fast --force {{input}}
		\"""
""")

	def write_roary(self):
		return(f"""
rule run_roary:
	input:
		expand("prokka/{{sample}}/{{sample}}.gff", sample = SAMPLE)
	output:
		"roary/gene_presence_absence.csv", "roary/summary_statistics.txt"
	threads:
		36
	shell:
		\"""
		roary -p {{threads}} -f roary {{input}}
		mv roary_*/* roary
		rm -r roary_*
		\"""
""")

	def write_pan_graph(self):
		return(f"""
rule pan_figure:
	input:
		"roary/gene_presence_absence.csv"
	output:
		"pan_genome.svg"
	shell:
		\"""
		roary2svg.pl {{input}} > {{output}}
		\"""
""")

	def write_assembly_stats(self):
		return(f"""
rule assembly_statistics:
	input:
		expand("{{sample}}/{{sample}}.fa", sample = SAMPLE)
	output:
		"denovo.tab"
	shell:
		\"""
		fa --t --each --minsize=500 {{input}} | sed 's@.*/@@' | sed 's/\.fa//g' > {{output}}
		\"""
	
""")

	def write_gff_summary(self):
		return(f"""
rule combine_assembly_metrics:
	input:
		prokka = expand("prokka/{{sample}}/{{sample}}.txt",sample = SAMPLE), 
		assembly = "denovo.tab"
	output:
		"assembly.tab"
	run:
		import pandas, pathlib

		prokka = f"{{input.prokka}}".split()
		gff = pandas.DataFrame()
		
		for p in prokka:
			g = pathlib.Path(p)
			df = pandas.read_csv(g, sep = ':', header = None, names = ['cond', f"{{g.parts[1]}}"])
			
			if gff.empty:
					gff = df
			else:
					gff = gff.merge(df, how = 'outer')
		gff = gff[gff['cond'].isin(['CDS', 'rRNA'])]
		gff = gff.T
		gff.columns = gff.iloc[0]
		gff = gff.iloc[1:]
	
		d = pathlib.Path(f"{{input.assembly}}")
		df = pandas.read_csv(d, sep = '\\t')
	
		assembly = df.merge(gff, left_on = ['Name'], right_on= gff.index)
		assembly = assembly.rename(columns={{'Name':'Isolate'}})
		assembly.to_csv(f"{{output}}", sep = '\\t', index = False)


""")
	def io_collation(self):
		i = f"'seqdata.tab'"
		o = f"'report/seqdata.tab'"
		# snps input string
		si = f"'core.txt', 'core.treefile', 'core.tab', 'distances.tab', 'core.tab'"
		# snps output string
		so = f"'report/core_genome.tab', 'report/core.treefile','report/distances.tab','report/core.tab'"
		# assembly input string
		ai = f"'assembly.tab', 'mlst.tab', 'species_identification.tab', 'resistome.tab'"
		# assembly output string
		ao = f"'report/assembly.tab', 'report/mlst.tab', 'report/species_identification.tab', 'report/resistome.tab'"
		# roary input
		ri = f"'pan_genome.svg', 'roary/summary_statistics.txt'"
		# roary output 
		ro = f"'report/pan_genome.svg', 'report/summary_statistics.txt'"
		# input string based on pipeline value
		return(i,o,si,so,ai,ao,ri,ro)
	def run_setup(self):
		
		run_string = f"""
		import pandas, pathlib, subprocess, numpy
"""
		coretxt = f"""
		
		# for core.txt
		df = pandas.read_csv(pathlib.Path(f"core.txt"), sep = '\\t')
		df['% USED'] = 100 * (df['LENGTH'] - df['UNALIGNED'])/ df['LENGTH']
		df['% USED'] = df['% USED'].round(2)
		df = df.rename(columns={{'ID':'Isolate'}})
		df.to_csv(f"report/core_genome.tab", sep='\\t', index = False)

"""
		assemblystatstring = f"""

		# calculate mean + 2SD and use as cutoff for quality of contigs and fix column names
		dfass = pandas.read_csv(pathlib.Path(f"assembly.tab"), sep = '\\t')
		cut = dfass['no'].mean() + (2* dfass['no'].std())
		dfass['Quality'] = numpy.where(dfass['no'] <= cut, 'PASS','FAIL')
		dfass = dfass.rename(columns={{'no':'# Contigs','Ns':'# Ns', 'gaps':'# Gaps', 'min':'Min Contig size', 'avg':'Avg Contig size', 'max':'Max Contig size', 'rRNA':'# rRNA', 'CDS':'# CDS'}})
		dfass.to_csv(f"report/assembly.tab", sep= '\\t', index=False)
"""
		cmdstring = f"""cp seqdata.tab report/seqdata.tab
"""
		scmd = f"""cp core.treefile report/core.treefile
cp distances.tab report/distances.tab
cp core.tab report/core.tab
"""
		acmd = f"""cp mlst.tab report/mlst.tab
cp species_identification.tab report/species_identification.tab
cp resistome.tab report/resistome.tab
"""
		rcmd = f"""cp pan_genome.svg report/pan_genome.svg
cp roary/summary_statistics.txt report/summary_statistics.txt
"""
		cmd = f"""
"""
		r_dict = {
			's' : f"""
		{run_string}
		{coretxt}
		cmd = f\"""
{cmdstring}{scmd}
		\"""
		subprocess.run(cmd, shell = True)
""",
			'a' : f"""
		{run_string}
		{assemblystatstring}
		cmd = f\"""
{cmdstring}{acmd}
		\"""
		subprocess.run(cmd, shell = True)
""",
			'sa' : f"""
		{run_string}
		{coretxt}
		{assemblystatstring}
		cmd = f\"""
{cmdstring}{scmd}{acmd}
		\"""
		subprocess.run(cmd, shell = True)
""",
			'all' : f"""
		{run_string}
		{coretxt}
		{assemblystatstring}
		cmd = f\"""
{cmdstring}{scmd}{acmd}{rcmd}
		\"""
		subprocess.run(cmd, shell = True)
"""}
		return(r_dict)


	def write_report_collation(self, pipeline = 'sa'):
		# input string always starts with seqdata and output string always starts with seqdata
		i,o,si,so,ai,ao,ri,ro = self.io_collation()
		
		io_dict = {
			's' : {
				'input':f"{i}, {si}",
				'output':f"{o}, {so}"
			},
			'a' : {
				'input':f"{i}, {ai}",
				'output':f"{o}, {ao}"
			},
			'sa' : {
				'input':f"{i}, {ai}, {si}",
				'output':f"{o}, {ao}, {so}"
			},
			'all' : {
				'input':f"{i}, {ai}, {si}, {ri}",
				'output':f"{o}, {ao}, {so}, {ro}"
			}
		}
		r_dict = self.run_setup()
		input_string = io_dict[pipeline]['input']
		output_string = io_dict[pipeline]['output']
		run_string = r_dict[pipeline]
		return(f"""
rule collate_report:
	input:
		{input_string}
	output:
		{output_string}
	run:
		{run_string}
""")

	def write_html(self, workdir, resources, job_id, script_path,assembler, pipeline = 'sa'):
		seq = f"'report/seqdata.tab'"
		s = f"'report/core_genome.tab', 'report/core.treefile', 'report/distances.tab'"
		a = f"'report/assembly.tab','report/mlst.tab', 'report/species_identification.tab','report/resistome.tab'"
		r = f"'report/pan_genome.svg', 'report/summary_statistics.txt'"
		if pipeline == 's':
			input_string = f"{seq},{s}"
		elif pipeline == 'a':
			input_string = f"{seq}, {a}"
		elif pipeline == 'sa':
						input_string = f"{seq}, {a}, {s}"
		elif pipeline == 'full':
			input_string = f"{seq}, {a}, {s}, {r}"
		wd_string = f"{workdir / job_id}"
		return(f"""

rule write_html_report:
	input:
		{input_string}
	output:
		'report/report.html'
	singularity:
		"shub://phgenomics-singularity/ete3@latest"
	shell:
		\"""
		python3 {script_path}/write_report.py {wd_string} {resources} {pipeline} {job_id} {assembler}
		\"""

		
		

""")
	
	def write_gubbins(self, gubbins,job):
		if gubbins == True:
			return("""
rule run_gubbins:
		input:
			'core.full.aln',
			'core.treefile'
		
		output:
			'clean.full.aln', 
			'gubbins.aln'
			
		# TODO fix gubbins!!
		shell:
			\"""
			snippy-clean_full_aln {{input[0]}} > {{output[0]}}
			run_gubbins.py -c 36 --starting_tree {{input[1]}} {{output[0]}}
			snp-sites -c clean.full.filtered_polymorphic_sites.fasta > {{output[1]}}
			\"""
""")
		else:
			return('')
	



from setuptools import setup, find_packages
import bohra


with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = bohra.__version__

setup (name = 'bohra',
       version = VERSION,
       classifiers = [
         "Programming Language :: Python :: 3.6",
         "Programming Language :: Python :: 3.7",
         "Operating System :: OS Independent",
         "Development Status :: 4 - Beta ",
         "Intended Audience :: Science/Research",
         "Topic :: Scientific/Engineering :: Bio-Informatics",
         "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
         ],
       include_package_data = True,
       packages=['bohra'],
       description = 'A bioinformatics pipeline for analysing short read Illumina data microbiological public health.',
       long_description = long_description,
       long_description_content_type="text/markdown",
       author = 'Kristy Horan',
       url = 'https://github.com/kristyhoran/bohra',
       install_requires = ['jinja2','biopython>=1.70','pandas>=0.23.0', 'numpy', 'svgwrite', 'psutil'],
       python_requires='>=3.6',
       entry_points={
        "console_scripts": [
           "bohra=bohra.bohra:main"
        ],},
        setup_requires=['pytest-runner'],
        tests_require = ['pytest'],
        test_suite = 'test'
)



#!/usr/bin/env python3



"""Bohra 
.. moduleauthor:: Kristy Horan <kristyhoran15@gmail.com>

Bohra exinct tree kangaroo that lived on the nullarbor

Bohra is microbial genomics pipeline, designed predominantly for use in public health, but may also be useful in research settings. The pipeline takes as input a tab-delimited file with the isolate IDs followed by the path to READ1 and READ2, a reference for alignment and a unique identifier, where reads are illumina paired end reads (other platforms are not supported).

The pipline is based on nullarbor (https://github.com/tseemann/nullarbor) and is designed to be run in high performance computing environment.

Bohra is modular allowing the user to choose between calling SNPs and generating a phylogenetic tree, perform assemblies and detect AMR, perform typing etc; Or use the full pipeline to call SNPs, generate phylogenies, assemble, type and detect the pan-geneome. The output of Bohra is a html report that can be distributed, with downloadable tables and data.
"""


import logging
import argparse
import pathlib
import sys
import os
from bohra.SnpDetection import RunSnpDetection
from bohra.ReRunSnpDetection import ReRunSnpDetection



#logging.basicConfig(filename='job.log',level=logging.INFO, format='[%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')

def run_pipeline(args):
    '''
    Run the pipeline for the first time
    '''
    R = RunSnpDetection(args)
    return(R.run_pipeline())

def rerun_pipeline(args):
    '''
    Rerun the pipeline on a previous dataset, adding, removing isolates or changing reference or mask file
    '''
    R = ReRunSnpDetection(args)
    return(R.run_pipeline())


def set_parsers():
    # setup the parser
  
    parser = argparse.ArgumentParser(description='Bohra - a bacterial genomics pipeline',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # subparser for running the pipeline
    subparsers = parser.add_subparsers(help="Task to perform")

    parser_sub_run = subparsers.add_parser('run', help='Initial run of Bohra', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    
    # options for running
    parser_sub_run.add_argument('--input_file','-i',help='Input file = tab-delimited with 3 columns <isolatename>  <path_to_read1> <path_to_read2>', default='')
    parser_sub_run.add_argument('--job_id','-j',help='Job ID, will be the name of the output directory', default='')
    parser_sub_run.add_argument('--reference','-r',help='Path to reference (.gbk or .fa)', default = '')
    parser_sub_run.add_argument('--mask','-m',default = False, help='Path to mask file if used (.bed)')
    parser_sub_run.add_argument('--pipeline','-p', default = 'sa', choices=['sa','s','a', 'all'], help=f"The pipeline to run. SNPS ('s') will call SNPs and generate phylogeny, ASSEMBLIES ('a') will generate assemblies and perform mlst and species identification using kraken2, SNPs and ASSEMBLIES ('sa' - default) will perform SNPs and ASSEMBLIES. ALL ('all') will perform SNPS, ASSEMBLIES and ROARY for pan-genome analysis")
    parser_sub_run.add_argument('--assembler','-a', default = 'shovill', choices=['shovill','skesa','spades'], help=f"Assembler to use.")
    parser_sub_run.add_argument('--cpus','-c',help='Number of CPU cores to run, will define how many rules are run at a time', default=36)
    parser_sub_run.add_argument('--minaln','-ma',help='Minimum percent alignment', default=0)
    parser_sub_run.add_argument('--prefillpath','-pf',help='Path to existing assemblies - in the form path_to_somewhere/isolatename/contigs.fa')
    parser_sub_run.add_argument('--mdu', default=True, help='If running on MDU data')
    parser_sub_run.add_argument('--workdir','-w', default = f"{pathlib.Path.cwd().absolute()}", help='Working directory, default is current directory')
    parser_sub_run.add_argument('--resources','-s', default = f"{pathlib.Path(__file__).parent / 'templates'}", help='Directory where templates are stored')
    parser_sub_run.add_argument('--force','-f', action="store_true", help = "Add if you would like to force a complete restart of the pipeline. All previous logs will be lost.")
    parser_sub_run.add_argument('--dryrun','-n', action="store_true", help = "If you would like to see a dry run of commands to be executed.")
    parser_sub_run.add_argument('--gubbins','-g', action="store_true", help = "If you would like to run gubbins. NOT IN USE YET - PLEASE DO NOT USE")
    # parser for rerun
    
    parser_sub_rerun = subparsers.add_parser('rerun', help='Rerun of Bohra. Add or remove isolates from isolate list, change mask or reference.', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # options for rerun
    parser_sub_rerun.add_argument('--reference','-r',help='Path to reference (.gbk or .fa)', default = '')
    parser_sub_rerun.add_argument('--mask','-m',default = '', help='Path to mask file if used (.bed)')
    parser_sub_rerun.add_argument('--cpus','-c',help='Number of CPU cores to run, will define how many rules are run at a time', default=36)
    parser_sub_rerun.add_argument('--workdir','-w', default = f"{pathlib.Path.cwd().absolute()}", help='Working directory, default is current directory')
    parser_sub_rerun.add_argument('--resources','-s', default = f"{pathlib.Path(__file__).parent / 'templates'}", help='Directory where templates are stored')
    parser_sub_rerun.add_argument('--dryrun','-n', action="store_true", help = "If you would like to see a dry run of commands to be executed.")
    parser_sub_rerun.add_argument('--gubbins','-g', action="store_true", help = "If you would like to run gubbins. NOT IN USE YET - PLEASE DO NOT USE")
    parser_sub_rerun.add_argument('--keep', '-k', action= 'store_true', help="Keep report from previous run")
    
    
    parser_sub_run.set_defaults(func=run_pipeline)
    
    parser_sub_rerun.set_defaults(func = rerun_pipeline)
    args = parser.parse_args()
    return(args)

def main():
    '''
    Run Bohra
    '''
    args = set_parsers()
    if vars(args) == {}:
        parser.print_help(sys.stderr)
    else:
        logging.basicConfig(filename='job.log',level=logging.INFO, format='[%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
        logger = logging.getLogger(__name__)
        args.func(args)
	
if __name__ == '__main__':
    main()

